﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Vintagestory.API.Client;
using Vintagestory.API.Server;
using Vintagestory.API.Common;
using Vintagestory.GameContent;
using Vintagestory.API.Config;
using Vintagestory.API.MathTools;
using Vintagestory.API.Common.Entities;
using ProtoBuf;

namespace vsmovemod.src
{
    // ProtoBuf
    [ProtoContract(ImplicitFields = ImplicitFields.AllPublic)]
    public class NetworkApiTestMessage
    {
        public bool messagefromserver;
        public bool sprintStep;
        public bool walkStep;
        public bool noStep;
    }

    [ProtoContract(ImplicitFields = ImplicitFields.AllPublic)]
    public class NetworkApiTestResponse
    {
        public bool response;
    }
    // ProtoBuf

    public class MoveMod : ModSystem
    {
        private ICoreServerAPI api;
        private ICoreClientAPI capi;

        public static IServerNetworkChannel serverChannel;
        public static IClientNetworkChannel clientChannel;

        // Item Init
        public override void Start(ICoreAPI api)
        {
            base.Start(api);
            api.RegisterItemClass("ItemGlider", typeof(ItemGlider));
        }


        // SERVERSIDE =======================================================================================================================================================================================================================================
        public override void StartServerSide(ICoreServerAPI api)
        {
            this.api = api;

            api.Event.PlayerJoin += entity =>
            entity.Entity.Properties.FallDamage = true;

            serverChannel =
                api.Network.RegisterChannel("networkToggleGlider")
                .RegisterMessageType(typeof(NetworkApiTestMessage))
                .RegisterMessageType(typeof(NetworkApiTestResponse))
                .SetMessageHandler<NetworkApiTestResponse>(OnClientMessage)
            ;

            // Register Commands
            api.RegisterCommand("autostep", "Send a test network message", "", OnNwTestCmd, Privilege.controlserver);
            api.RegisterCommand("sprint-step", "Send a test network message", "", SprintStep, Privilege.controlserver);
            api.RegisterCommand("walk-step", "Send a test network message", "", WalkStep, Privilege.controlserver);
            api.RegisterCommand("no-step", "Send a test network message", "", NoStep, Privilege.controlserver);
        }

        // Command Methods (choosing stepheight)
        private void SprintStep(IServerPlayer player, int groupId, CmdArgs args)
        {
            serverChannel.BroadcastPacket(new NetworkApiTestMessage()
            {
                messagefromserver = false,
                sprintStep = true,
                walkStep = false,
                noStep = false,
        });
        }
        private void WalkStep(IServerPlayer player, int groupId, CmdArgs args)
        {
            serverChannel.BroadcastPacket(new NetworkApiTestMessage()
            {
                messagefromserver = false,
                sprintStep = false,
                walkStep = true,
                noStep = false,
        });
        }
        private void NoStep(IServerPlayer player, int groupId, CmdArgs args)
        {
            serverChannel.BroadcastPacket(new NetworkApiTestMessage()
            {
                messagefromserver = false,
                sprintStep = false,
                walkStep = false,
                noStep = true,
        });
        }
        private void OnNwTestCmd(IServerPlayer player, int groupId, CmdArgs args)
        {
            serverChannel.BroadcastPacket(new NetworkApiTestMessage()
            {
                messagefromserver = false,
            });
        }

        private void OnClientMessage(IPlayer fromPlayer, NetworkApiTestResponse networkMessage)
        {
            fromPlayer.Entity.Properties.FallDamage = networkMessage.response;

            //api.SendMessageToGroup(
            //    GlobalConstants.GeneralChatGroup,
            //    "Network toggled falldamage to: " + fromPlayer.Entity.Properties.FallDamage.ToString() + " Response: " + networkMessage.response.ToString(),
            //    EnumChatType.Notification
            //);
        }

        // Action on Server-Command (setting autostep-mode)
        private void OnServerMessage(NetworkApiTestMessage networkMessage)
        {
            if (networkMessage.sprintStep)
            {
                DataFields.sprintStep = true;
                DataFields.walkStep = false;
                DataFields.noStep = false;

                capi.ShowChatMessage("Autostep set to sprint");
            }
            else if (networkMessage.walkStep)
            {
                DataFields.sprintStep = false;
                DataFields.walkStep = true;
                DataFields.noStep = false;
                capi.ShowChatMessage("Autostep set to walk");
            }
            else if (networkMessage.noStep)
            {
                DataFields.sprintStep = false;
                DataFields.walkStep = false;
                DataFields.noStep = true;
                capi.ShowChatMessage("Autostep disabled");
            }
        }

        // CLIENTSIDE =======================================================================================================================================================================================================================================
        public override void StartClientSide(ICoreClientAPI api)
        {
            this.capi = api;

            clientChannel =
                api.Network.RegisterChannel("networkToggleGlider")
                .RegisterMessageType(typeof(NetworkApiTestMessage))
                .RegisterMessageType(typeof(NetworkApiTestResponse))
                .SetMessageHandler<NetworkApiTestMessage>(OnServerMessage)
            ;

            //Spawn Init - set stepheight
            if (Config.walkstep)
            {
                api.Event.PlayerEntitySpawn += entity =>
                entity.Entity.GetBehavior<EntityBehaviorControlledPhysics>().stepHeight = 1F;
            }
            else
            {
                api.Event.PlayerEntitySpawn += entity =>
                entity.Entity.GetBehavior<EntityBehaviorControlledPhysics>().stepHeight = 0.9F;
            }

            //Spawn Init - set climbing flase
            api.Event.PlayerEntitySpawn += entity =>
            entity.Entity.Properties.CanClimbAnywhere = false;

            // Register HotKey Stop crawling
            api.Input.RegisterHotKey("standup", "Stand Up from crawling", GlKeys.Space, HotkeyType.CharacterControls);
            api.Input.SetHotKeyHandler("standup", StandUp);

            // Register HotKey Start Crawl
            api.Input.RegisterHotKey("startcrawl", "Crawl on the floor using Sneak + Hotkey", GlKeys.R, HotkeyType.CharacterControls);
            api.Input.SetHotKeyHandler("startcrawl", StartCrawl);

            // Register HotKey Use Glider
            api.Input.RegisterHotKey("useglider", "Use Glider when equiped / Let-go when climbing", GlKeys.F, HotkeyType.CharacterControls);
            api.Input.SetHotKeyHandler("useglider", UseGlider);
          
            // Game Time Tick Listener 1ms ========================================================
            api.Event.RegisterGameTickListener(dt2 =>
            {
                var player = api.World.Player;

                // Stop Animation when inGliderMode
                ItemGlider.StopAnimations(player);

                // Prevent player from climbing down (while bouldering)
                BehaviorClimbing.DisableClimbingDown(player, api);

                // Stop Motion Vec AfterDash
                BehaviorDash.DashStop(player, api);

                // Toggle fall-damage when gliding / climbing
                BehaviorFalldamage.SetFallDamage(player, api);

            }, 1);
           
            // Game Time Tick Listener 10ms ========================================================
            api.Event.RegisterGameTickListener(dt1 =>
            {
                var player = api.World.Player;

                // Glider stuff
                ItemGlider.UseGlider(player, api);

            }, 10); //10 ms/tick

            // Game Time Tick Listener 50ms ========================================================
            api.Event.RegisterGameTickListener(dt =>
            {
                var player = api.World.Player;
                DataFields.clientUID = player.PlayerUID;

                // Get block at playerPos
                Block checkBlockatPlayerPos = api.World.BlockAccessor.GetBlock((int)player.Entity.Pos.X, (int)player.Entity.Pos.Y, (int)player.Entity.Pos.Z);

                // Check if player is climbing for GUI
                if (player.Entity.Controls.IsClimbing)
                {
                    DataFields.playerIsClimbing = true;
                }
                else
                {
                    DataFields.playerIsClimbing = false;
                }

                // Check if block at playerPos is climable (ladder)
                if (checkBlockatPlayerPos.Climbable)
                {
                    DataFields.isClimbable = true;
                }
                else
                {
                    DataFields.isClimbable = false;
                }

                // Debug Outputs
                Debug.OutputDebug(player);

                // Timer Method here
                Timers.HandleCounters();

                // Hangclimbing Method here
                BehaviorHangclimbing.Hanging(player, api);

                // Crawling Method here
                BehaviorCrawling.Crawling(player, api);

                // Dash Method here 
                BehaviorDash.DashMovement(player, api);

                // Slide Method here 
                BehaviorSliding.Sliding(player, api);

                // Walljump Method here
                BehaviorWalljump.Walljump(player);

                // Chargejump Method here
                BehaviorChargejump.Chargejump(player);
      
                // Climb Method here 
                BehaviorClimbing.EnableClimbing(player, api);

                //Stepheight Method here
                BehaviorStepheight.Stepheight(player);

                //Stamina Method here
                BehaviorStamina.CalculateStamina(player, api);

            }, 50); //50 ms/tick
        }

        // Hotkey stand-up Method
        public static bool StandUp(KeyCombination comb)
        {
            if (DataFields.isCrawling)
            {
                DataFields.toggleStandup = true;
                return true;
            }

            return true;
        }

        // Hotkey start-crawl Method
        public static bool StartCrawl(KeyCombination comb)
        {
            if (!DataFields.isCrawling)
            {
                DataFields.toggleCrawl = true;
                return true;
            }

            return true;
        }

        // Hotkey toggle-glider Method
        public static bool UseGlider(KeyCombination comb)
        {
            if (!DataFields.playerIsClimbing)
            {
                if (!DataFields.toggleGlider)
                {
                    DataFields.toggleGlider = true;
                    return true;
                }
                else if (DataFields.toggleGlider)
                {
                    DataFields.toggleGlider = false;
                    return true;
                }
            }
            else if (DataFields.playerIsClimbing)
            {
                DataFields.letGoClimbing = true;
                return true;
            }
            return true;
        }

        // EventHandler PlayerAction - Prevent Sneaking / climbing down
        public static void Input_InWorldAction3(EnumEntityAction action, bool on, ref EnumHandling handled)
        {
            //Prevent sneak
            if (action == EnumEntityAction.Sneak)
            {
                handled = EnumHandling.PreventDefault;
            }
        }

        // EventHandler PlayerAction - Prevent sneaking / Sprinting
        public static void Input_InWorldAction4(EnumEntityAction action, bool on, ref EnumHandling handled)
        {
            //Prevent sneak
            if (action == EnumEntityAction.Sneak)
            {
                handled = EnumHandling.PreventDefault;
            }
            //Prevent sprint
            if (action == EnumEntityAction.Sprint)
            {
                handled = EnumHandling.PreventDefault;
            }
        }

        // EventHandler PlayerAction - Allow all movement & Prevent Sprint,Jump
        public static void Input_InWorldAction2(EnumEntityAction action, bool on, ref EnumHandling handled)
        {
            //Prevent sprint / jump
            if (action == EnumEntityAction.Sprint)
            {
                handled = EnumHandling.PreventDefault;
            }
            else if (action == EnumEntityAction.Jump)
            {
                handled = EnumHandling.PreventDefault;
            }

            //allow all movement directions
            else if (action == EnumEntityAction.Forward)
            {
                handled = EnumHandling.PassThrough;
            }
            else if (action == EnumEntityAction.Left)
            {
                handled = EnumHandling.PassThrough;
            }
            if (action == EnumEntityAction.Right)
            {
                handled = EnumHandling.PassThrough;
            }
            else if (action == EnumEntityAction.Backward)
            {
                handled = EnumHandling.PassThrough;
            }
        }

        // EventHandler PlayerAction - Prevent all directions & Sprint,Jump
        public static void Input_InWorldAction(EnumEntityAction action, bool on, ref EnumHandling handled)
        {
            //Prevent sprint / jump
            if (action == EnumEntityAction.Sprint)
            {
                handled = EnumHandling.PreventDefault;
            }
            else if (action == EnumEntityAction.Jump)
            {
                handled = EnumHandling.PreventDefault;
            }

            //prevent all movement directions
            if (action == EnumEntityAction.Forward)
            {
                handled = EnumHandling.PreventDefault;
            }
            else if (action == EnumEntityAction.Left)
            {
                handled = EnumHandling.PreventDefault;
            }
            if (action == EnumEntityAction.Right)
            {
                handled = EnumHandling.PreventDefault;
            }
            else if (action == EnumEntityAction.Backward)
            {
                handled = EnumHandling.PreventDefault;
            }

        }

        // EventHandler PlayerAction - Allow all movement
        public static void Input_InWorldAction1(EnumEntityAction action, bool on, ref EnumHandling handled)
        {
            if (action == EnumEntityAction.Sprint)
            {
                handled = EnumHandling.PassThrough;
            }
            else if (action == EnumEntityAction.Jump)
            {
                handled = EnumHandling.PassThrough;
            }
            else if (action == EnumEntityAction.Sneak)
            {
                handled = EnumHandling.PassThrough;
            }

            //allow all movement directions
            else if (action == EnumEntityAction.Forward)
            {
                handled = EnumHandling.PassThrough;
            }
            else if (action == EnumEntityAction.Left)
            {
                handled = EnumHandling.PassThrough;
            }
            else if (action == EnumEntityAction.Right)
            {
                handled = EnumHandling.PassThrough;
            }
            else if (action == EnumEntityAction.Backward)
            {
                handled = EnumHandling.PassThrough;
            }
        }
    }

    // Stamina GUI
    // -----------------------------------------------------------------------------------
    public class HudOverlaySample : ModSystem
    {
        ICoreClientAPI capi;
        StaminaProgressBarRenderer renderer;

        public override bool ShouldLoad(EnumAppSide side)
        {
            return side == EnumAppSide.Client;
        }

        public override void StartClientSide(ICoreClientAPI api)
        {
            this.capi = api;
            renderer = new StaminaProgressBarRenderer(api);

            api.Event.RegisterRenderer(renderer, EnumRenderStage.Ortho);
        }
    }

    public class StaminaProgressBarRenderer : IRenderer
    {
        MeshRef whiteRectangleRef;
        MeshRef progressQuadRef;
        ICoreClientAPI capi;

        Matrixf mvMatrix = new Matrixf();

        public double RenderOrder { get { return 0; } }

        public int RenderRange { get { return 10; } }

        public StaminaProgressBarRenderer(ICoreClientAPI api)
        {
            this.capi = api;

            // This will get a line loop with vertices inside [-1,-1] till [1,1]
            MeshData rectangle = LineMeshUtil.GetRectangle(ColorUtil.WhiteArgb);
            whiteRectangleRef = api.Render.UploadMesh(rectangle);

            // This will get a quad with vertices inside [-1,-1] till [1,1]
            progressQuadRef = api.Render.UploadMesh(QuadMeshUtil.GetQuad());
        }

        public void OnRenderFrame(float deltaTime, EnumRenderStage stage)
        {
            IShaderProgram curShader = capi.Render.CurrentActiveShader;

            Vec4f color = new Vec4f(1, 1, 1, 1);

            if (DataFields.playerIsClimbing == true || DataFields.stamina < DataFields.maxStamina)
            {
                // Render rectangle
                curShader.Uniform("rgbaIn", color);
                curShader.Uniform("extraGlow", 0);
                curShader.Uniform("applyColor", 0);
                curShader.Uniform("tex2d", 0);
                curShader.Uniform("noTexture", 1f);

                float width = 100;

                mvMatrix
                    .Set(capi.Render.CurrentModelviewMatrix)
                    .Translate(capi.Render.FrameWidth / 2 - 50, capi.Render.FrameHeight / 8, 0)
                    .Scale(width, 20, 0)
                    .Translate(0.5f, 0.5f, 0)
                    .Scale(0.5f, 0.5f, 0)
                ;
                curShader.UniformMatrix("projectionMatrix", capi.Render.CurrentProjectionMatrix);
                curShader.UniformMatrix("modelViewMatrix", mvMatrix.Values);

                capi.Render.RenderMesh(whiteRectangleRef);
            }


            // Render progress bar
            if (DataFields.playerIsClimbing == true || DataFields.stamina < DataFields.maxStamina)
            {
                if (DataFields.stamina >= DataFields.maxStamina)
                {
                    float width = 100;
                    mvMatrix
                        .Set(capi.Render.CurrentModelviewMatrix)
                        .Translate(capi.Render.FrameWidth / 2 - 50, capi.Render.FrameHeight / 8, 0)
                        .Scale(width, 20, 0)
                        .Translate(0.5f, 0.5f, 0)
                        .Scale(0.5f, 0.5f, 0)
                    ;
                }
                else
                {
                    float width = ((DataFields.stamina) / DataFields.maxStamina) * 100;
                    mvMatrix
                        .Set(capi.Render.CurrentModelviewMatrix)
                        .Translate(capi.Render.FrameWidth / 2 - 50, capi.Render.FrameHeight / 8, 0)
                        .Scale(width, 20, 0)
                        .Translate(0.5f, 0.5f, 0)
                        .Scale(0.5f, 0.5f, 0)
                    ;
                }
            }

            if (DataFields.playerIsClimbing == true || DataFields.stamina < DataFields.maxStamina)
            {
                curShader.UniformMatrix("projectionMatrix", capi.Render.CurrentProjectionMatrix);
                curShader.UniformMatrix("modelViewMatrix", mvMatrix.Values);

                capi.Render.RenderMesh(progressQuadRef);
            }

        }

        public void Dispose()
        {
            capi.Render.DeleteMesh(whiteRectangleRef);
            capi.Render.DeleteMesh(progressQuadRef);
        }
    }
}
